<?php
/**
 * @file
 * An exclusive-OR encryption provider. This class provides the ability to use
 * exclusive-OR (XOR) operations and random padding to encrypt and decrypt
 * values.
 *
 * When encrypting a value, the value is randomly padded and then
 * exclusively-OR-ed with an encryption key to produce the encrypted value. When
 * decrypting an encrypted value, the encrypted value is exclusively-OR-ed with
 * the same encryption key, and then stripped of the padding digits, to produce
 * the original, un-encrypted value.
 *
 * During either process, if a value has a longer bit-length than the encryption
 * key, the encryption key is automatically expanded to match the bit-length of
 * the value.
 *
 * @author Guy Paddock (guy.paddock@redbottledesign.com)
 */
class OplEncryptor {
  /**
   * The default "base" or radix of numbers (i.e. base-10, base-16, etc) that
   * will be encrypted.
   */
  const DEFAULT_RADIX = 10;

  /**
   * The default number of random padding digits to add to an encrypted string
   * to make it more difficult to derive the key.
   */
  const DEFAULT_NUM_PAD_DIGITS = 3;

  /**
   * The "base" or radix of numbers (i.e. base-10, base-16, etc) that will be
   * encrypted.
   *
   * @var number
   */
  private $radix;

  /**
   * The number of random padding digits to add to an encrypted string to make
   * it more difficult to derive the key.
   *
   * @var integer
   */
  private $numPadDigits;

  /**
   * The key to use for encryption.
   *
   * @var number
   */
  private $key;

  /**
   * Initialize a new instance of OplEncryptor for the specified number base /
   * radix, number of padding digits, and encryption key.
   *
   * @param integer $radix
   *   The "base" or radix of numbers (i.e. base-10, base-16, etc) that will be
   *   encrypted. If unset, the default of DEFAULT_RADIX will be a used.
   * @param interger $numPadDigits
   *   The number of random padding digits to add to an encrypted string to make
   *   it more difficult to derive the key. If unset, the default of
   *   DEFAULT_NUM_PAD_DIGITS will be a used.
   * @param number $key
   *   The key to use for encryption. If unset, a default of (PI * e) ^ 10 will
   *   be used.
   */
  public function OplEncryptor($radix = OplEncryptor::DEFAULT_RADIX,
      $numPadDigits = OplEncryptor::DEFAULT_NUM_PAD_DIGITS,
      $key = NULL) {
    $this->radix = $radix;
    $this->numPadDigits = $numPadDigits;

    // Provide default key, if necessary.
    if ($key === NULL) {
      $this->key = gmp_init((int)pow(pi() * exp(1), 10), 10);
    }
    else {
      $this->key = gmp_init($key, $this->radix);
    }
  }

  /**
   * Encrypt the specified value.
   *
   * @param number $value
   *   The value to encrypt.
   *
   * @return string
   *   The encrypted form of the value.
   */
  public function encrypt($value) {
    // Convert to GMP resource.
    $resourceValue = gmp_init($value, $this->radix);

    // Add some random padding digits.
    $padded = $this->addRandomPadding($resourceValue);

    // Expand the key to match the bit length of the value.
    $expandedKey = $this->calculateExpandedKeyFor($padded);

    // Perform the encryption.
    $encrypted = gmp_xor($padded, $expandedKey);

    // Return the result as a string.
    return gmp_strval($encrypted, $this->radix);
  }

  /**
   * Decrypt the specified value.
   *
   * @param string $value
   *   The value to decrypt.
   *
   * @return number
   *   The decrypted form of the value.
   */
  public function decrypt($value) {
    // Convert to GMP resource.
    $resourceValue = gmp_init($value, $this->radix);

    // Expand the key to match the bit length of the value.
    $expandedKey = $this->calculateExpandedKeyFor($resourceValue);

    // Decrypt.
    $decrypted = gmp_xor($resourceValue, $expandedKey);

    // Remove the random padding digits.
    $withoutPadding = $this->removeRandomPadding($decrypted);

    // Return the result as a string.
    return gmp_strval($withoutPadding, $this->radix);
  }

  /**
   * Add random padding digits to the specified value.
   *
   * @param resource $value
   *   The GMP resource that represents the value to which random padding will
   *   be added.
   *
   * @return resource
   *   The GMP resource that represents $value with random padding added.
   */
  private function addRandomPadding($value) {
    $temp  = '';

    for ($i = 0; $i < $this->numPadDigits; ++$i) {
      /* Pick a random value in the range of the radix and convert it to a GMP
       * resource.
       */
      $randomDigit = gmp_init(rand(1, $this->radix - 1), 10);

      /* Convert the random digit to the appropriate base and append it to the
       * output string.
       */
      $temp .= gmp_strval($randomDigit, $this->radix);
    }

    // Now, append the original value.
    $temp .= gmp_strval($value, $this->radix);

    // Return the result as a GMP resource in the appropriate base.
    return gmp_init($temp, $this->radix);
  }

  /**
   * Remove random padding digits from the specified value.
   *
   * @param resource $value
   *   The GMP resource that represents the value from which random padding will
   *   be removed.
   *
   * @return resource
   *   The GMP resource that represents $value with random padding removed.
   */
  private function removeRandomPadding($value) {
    // Convert the resource to the appropriate base.
    $withPad  = gmp_strval($value, $this->radix);
    $withoutPad = '0';

    if ($this->numPadDigits < strlen($withPad)) {
      $withoutPad = substr($withPad, $this->numPadDigits);
    }

    // Return the result as a GMP resource in the appropriate base.
    return gmp_init($withoutPad, $this->radix);
  }

  /**
   * Calculate a version of the current key that has been expanded to match the
   * length (in bits) of the specified value, for encrypting or decrypting
   * values longer than the original key.
   *
   * @param resource $value
   *   A GMP resource that represents the value that the key must match in
   *   bit-length.
   *
   * @return resource
   *   A GMP resource that represents the key value of the necessary bit length.
   */
  private function calculateExpandedKeyFor($value) {
    $valueBits        = gmp_strval($value, 2);
    $desiredBitLength = strlen($valueBits);

    $keyBits      = gmp_strval($this->key, 2);
    $keyBitLength = strlen($keyBits);

    /* The prefix bit is always the complement of the first bit of the value
     * we're encrypting, to ensure that when the encryption key and the value
     * are XORed, the resulting value will be of a predictable length.
     */
    $prefixBit = ($valueBits[0] == 0) ? '1' : '0';

    $prefixedKeyBits      = $prefixBit . $keyBits;
    $prefixedKeyBitLength = $keyBitLength + 1;

    $newKeyBits = $prefixedKeyBits;

    for ($bitIndex = 0; $bitIndex < ($desiredBitLength - $prefixedKeyBitLength); ++$bitIndex) {
      // Reverse the key during expansion.
      $newKeyBits .= $keyBits[$keyBitLength - ($bitIndex % $keyBitLength) - 1];
    }

    return gmp_init($newKeyBits, 2);
  }
}
